
find :: (Eq a) => a -> [a] -> Maybe Int
find y xs = findH y xs 0
          where
          findH _ [] _ = Nothing
          findH y (x:xs) i | (y == x) = Just i
                           | otherwise = findH y xs (i + 1)

everythingAfter :: (Eq a) => a -> [a] -> Maybe [a]
everythingAfter y xs = case (find y xs) of
                         Nothing -> Nothing
                         Just i  -> Just (drop (i + 1) xs)


-- we can define everythingAfter without using find
everythingAfter' :: (Eq a) => a -> [a] -> Maybe [a]
everythingAfter' _ [] = Nothing
everythingAfter' x (y:ys) | (x == y) = Just ys
                          | otherwise = everythingAfter' x ys



-- a Map data structure is a list of key-value pairs
type Map k v = [(k, v)]

-- compare this get function with the get methods from Java's Map interface
--   https://docs.oracle.com/javase/8/docs/api/java/util/Map.html#get-java.lang.Object-
--   https://docs.oracle.com/javase/8/docs/api/java/util/Map.html#getOrDefault-java.lang.Object-V-
get :: (Eq k) => k -> Map k v -> Maybe v
get _ [] = Nothing
get x ((key,value):kvs) | (x == key) = Just value
                        | otherwise = get x kvs


-- this Map data structure is a list of key-value pairs where the value may be missing
type Map2 k v = [(k, Maybe v)]

-- redo the get function for maps that can have missing values
-- notice the "double maybe", one maybe for possible missing values and one maybe for possibly not finding the key
get2 :: (Eq k) => k -> Map2 k v -> Maybe (Maybe v)
get2 _ [] = Nothing
get2 x ((key,value):kvs) | (x == key) = Just value
                         | otherwise = get2 x kvs


-- Create a new data type for use by the get function.
data GetResult a = KeyNotFound
                 | KeyHasNoValue
                 | Found a
                 deriving Show

-- Redo the get function using the new return type.
get3 :: (Eq k) => k -> Map2 k v -> GetResult v
get3 _ [] = KeyNotFound
get3 x ((key,value):kvs) | (x == key) = case value of
                                          Nothing -> KeyHasNoValue
                                          Just i  -> Found i
                         | otherwise = get3 x kvs



-- try to implement the putIfAbsent() method from Java's Map interface
--    https://docs.oracle.com/javase/8/docs/api/java/util/Map.html#putIfAbsent-K-V-
putIfAbsent :: (Eq k) => Map2 k v -> (k, Maybe v) -> Map2 k v
putIfAbsent map (key, value) = map



map1 = [(3, "cat"), (1, "dog"), (4, "apple"), (2, "pear")]
map2 = [(5, Nothing), (3, Just "cat"), (1, Just "dog"), (0, Nothing), (4, Just "apple"), (2, Just "pear")]
